#include "Creep.h"
#include "Player.h"
#include "Room.h"

const int Creep::HURT_TIME = 30;
const float Creep::WIDTH = 8;
const float Creep::HEIGHT = 8;
const float Creep::MAX_SPEED = 1.5;
const float Creep::MAX_FORCE = 0.02;
const float Creep::NEIGHBOR_RADIUS = 25;
const float Creep::CLOSE_NEIGHBOR_RADIUS = 15;

Creep::Creep(float x, float y):
    Entity(CREEP, x, y, WIDTH, HEIGHT),
    sprite(ResourceManager::getTexture("creep.png"), 16, 16, -1, 15, 0, 2, true),
    acceleration(0, 0),
    separation(0, 0),
    alignment(0, 0),
    cohesion(0, 0),
    neighborCount(0),
    closeNeighborCount(0),
    sleepTimer( 60 + rand()%30),
    hurtTimer(0)
{
    sprite.setRandomFrame();
    dx = (rand()/(float)RAND_MAX - 0.5f) / 10.f;
    dy = (rand()/(float)RAND_MAX - 0.5f) / 10.f;
    bouncy = true;
    hp = 2;
}

Creep::~Creep()
{
    //dtor
}

bool Creep::hurt(int dmg)
{
    if(hurtTimer > 0)
    {
        return false;
    }
    dx = 0;
    dy = 0;
    hurtTimer = HURT_TIME;
    bool wasHurt = Entity::hurt(dmg);
    if(wasHurt)
    {
        room->addParticle(new Particle(Particle::TYPE_BLOOD, x+w/2, y+h/2));
        room->addParticle(new Particle(Particle::TYPE_BLOOD, x+w/2, y+h/2));
        room->addParticle(new Particle(Particle::TYPE_BLOOD, x+w/2, y+h/2));
        room->addParticle(new Particle(Particle::TYPE_BLOOD, x+w/2, y+h/2));
        if(die)
        {
            ResourceManager::playSound("destroyed.wav");
            if(!room->findCreeps() && rand()%100 < 75)
            {
                room->addEntity(new Item(x,y,Item::ITEM_HEALTH));
            }
        }
        else
        {
            ResourceManager::playSound("hit.wav");
        }
    }
    return wasHurt;
}
void Creep::step()
{
    if(sleepTimer > 0)
    {
        sleepTimer--;
    }
    else
    {
        flock();
    }
    if(hurtTimer > 0)
    {
        hurtTimer--;
    }
    dx += acceleration.x;
    dy += acceleration.y;
    acceleration.x = 0;
    acceleration.y = 0;
    float speed = sqrtf(dx * dx + dy * dy);
    if(speed > MAX_SPEED)
    {
        dx *= MAX_SPEED / speed;
        dy *= MAX_SPEED / speed;
    }
    sprite.setSpeed(4 * MAX_SPEED / speed);
    if(fabs(dx) > 32 || fabs(dy) > 32)
    {
        printf("SUPER SPEED DETECTED: %f, %f\n", dx, dy);
    }
    x += dx;
    y += dy;
    collideRoom();
    if(landedLeft)
    {
        acceleration.y -= 0.05;
    }
    if(landedRight)
    {
        acceleration.y += 0.05;
    }
    if(landedUp)
    {
        acceleration.x -= 0.05;
    }
    if(landedDown)
    {
        acceleration.x += 0.05;
    }
    facingRight = dx > 0;
    sprite.step();

}

void Creep::draw(sf::RenderTarget& window)
{
    if(hurtTimer > 0)
    {
        float f = 1.f - (float)hurtTimer / HURT_TIME;
        sprite.setColor(sf::Color(255, 255 * f, 255 * f));
        sprite.setPosition(x + w/2 + rand()%3 - 1, y + h/2);
    }
    else
    {
       sprite.setColor(sf::Color(255, 255, 255));
        sprite.setPosition(x + w/2, y + h/2);
    }
    sprite.setFlipped(!facingRight);
    sprite.draw(window);
}

void Creep::doCollide(Entity* other)
{
    if(other->entityType == CREEP)
    {
        float dist = sqrtf(powf(other->x - x, 2) + powf(other->y - y, 2));
        if(dist <= 0)
        {
            return;
        }
        if(dist < NEIGHBOR_RADIUS)
        {
            neighborCount++;
            // Alignment
            alignment.x += other->dx;
            alignment.y += other->dy;
            // Cohesion
            cohesion.x += other->x;
            cohesion.y += other->y;

            if(dist < CLOSE_NEIGHBOR_RADIUS)
            {
                sf::Vector2f diff(x - other->x, y - other->y);
                float magDiff = sqrtf(diff.x * diff.x + diff.y * diff.y);
                diff.x /= magDiff * dist;
                diff.y /= magDiff * dist;
                separation += diff;
                closeNeighborCount++;
            }
        }
    }
    else if(other->entityType == PLAYER)
    {
        if(collidesWith(other))
        {
            other->hurt(1);
        }
    }
}

sf::Vector2f Creep::seek(sf::Vector2f target)
{
    sf::Vector2f desired = target - sf::Vector2f(x, y);
    float mag = sqrtf(desired.x * desired.x + desired.y * desired.y);
    desired.x *= MAX_SPEED / mag;
    desired.y *= MAX_SPEED / mag;

    desired.x -= dx;
    desired.y -= dy;
    //limit
    mag = sqrtf(desired.x * desired.x + desired.y * desired.y);
    if(mag > MAX_FORCE)
    {
        desired.x *= MAX_FORCE / mag;
        desired.y *= MAX_FORCE / mag;
    }
    return desired;
}
sf::Vector2f Creep::avoid(sf::Vector2f target)
{
    sf::Vector2f desired = sf::Vector2f(x, y) - target;
    float mag = sqrtf(desired.x * desired.x + desired.y * desired.y);
    desired.x *= MAX_SPEED / mag;
    desired.y *= MAX_SPEED / mag;

    desired.x -= dx;
    desired.y -= dy;
    //limit
    mag = sqrtf(desired.x * desired.x + desired.y * desired.y);
    if(mag > MAX_FORCE)
    {
        desired.x *= MAX_FORCE / mag;
        desired.y *= MAX_FORCE / mag;
    }
    return desired;
}

void Creep::flock()
{
    if(neighborCount > 0)
    {
        /// Alignment
        //average
        alignment.x /= neighborCount;
        alignment.y /= neighborCount;
        if(alignment.x != 0 || alignment.y != 0)
        {
            float alignMag = sqrtf(alignment.x * alignment.x + alignment.y * alignment.y);
            //normalize
            alignment.x *= MAX_SPEED / alignMag;
            alignment.y *= MAX_SPEED / alignMag;
            alignment.x -= dx;
            alignment.y -= dy;
            //limit
            alignMag = sqrtf(alignment.x * alignment.x + alignment.y * alignment.y);
            if(alignMag > MAX_FORCE)
            {
                alignment.x *= MAX_FORCE / alignMag;
                alignment.y *= MAX_FORCE / alignMag;
            }
        }

        /// Cohesion
        cohesion.x /= neighborCount;
        cohesion.y /= neighborCount;
        cohesion = seek(cohesion);
    }
    if(closeNeighborCount > 0)
    {
        /// Separation
        separation.x /= closeNeighborCount;
        separation.y /= closeNeighborCount;
        if(separation.x != 0 || separation.y != 0)
        {
            float sepMag = sqrtf(separation.x * separation.x + separation.y * separation.y);
            //normalize
            separation.x *= MAX_SPEED / sepMag;
            separation.y *= MAX_SPEED / sepMag;
            separation.x -= dx;
            separation.y -= dy;
            sepMag = sqrtf(separation.x * separation.x + separation.y * separation.y);
            if(sepMag > MAX_FORCE)
            {
                separation.x *= MAX_FORCE / sepMag;
                separation.y *= MAX_FORCE / sepMag;
            }
        }
    }
    separation.x *= 1.5; //weight
    separation.y *= 1.5;
    cohesion.x *= 0.6;
    cohesion.y *= 0.6;
    alignment.x *= 1.2;
    alignment.y *= 1.2;

    acceleration += separation + alignment + cohesion;
    Player* player = room->getPlayer();
    if(player)
    {
        if(hurtTimer == 0)
        {
            acceleration += seek(sf::Vector2f(player->x, player->y));
        }
        else
        {
            acceleration += avoid(sf::Vector2f(player->x, player->y));
        }
    }
    if(isOutOfBounds())
    {
        sf::Vector2f centerRoomForce = sf::Vector2f(room->w/2 * Room::TILE_SIZE,
                                                         room->h/2 * Room::TILE_SIZE);
        centerRoomForce.x -= x;
        centerRoomForce.y -= y;

        dx = centerRoomForce.x;
        dy = centerRoomForce.y;
    }
    closeNeighborCount = 0;
    neighborCount = 0;
    alignment = sf::Vector2f(0, 0);
    separation = sf::Vector2f(0, 0);
    cohesion = sf::Vector2f(0, 0);
}
