#include "Player.h"
#include "Room.h"
#include "CombatParticle.h"

const int Player::HURT_TIME = 90;
const float Player::WIDTH = 14;
const float Player::HEIGHT = 14;
const float Player::MOVE_SPEED = 2;
const unsigned Player::MAX_ATTACK_FRAMES = 30;
const unsigned Player::FALL_FRAMES = 30;

Player::Player(float x, float y):
    Entity(PLAYER, x, y, WIDTH, HEIGHT),
    idleSprite(ResourceManager::getTexture("player.png"), 32, 32, -1, 15, 0, 0, true),
    fallSprite(ResourceManager::getTexture("player_fall.png"), 32, 32, -1, 15, 0, 0, true),
    deadSprite(ResourceManager::getTexture("player_dead.png"), 32, 32, -1, 15, 0, 0, true),
    walkSprite(ResourceManager::getTexture("player_walk.png"), 32, 32, -1, 5, 0, 1, true),
    attackSprite(ResourceManager::getTexture("player_attack.png"), 32, 32, -1, 1, 0, 1, false),
    currSprite(&idleSprite),
    statSprite(ResourceManager::getTexture("items.png"), 16, 16, -1, -1, 0, 0, false),
    facingRight(false),
    attacking(false),
    direction(1, 0),
    hurtTimer(0),
    running(false),
    moveSpeed(MOVE_SPEED),
    attackFrames(0),
    fallingIntoHole(false),
    fallTimer(0),
    attackPow(1),
    attackSpeed(1),
    runSpeed(1),
    numKeys(0),
    bossKey(false),
    descending(false),
    descentPercent(0),
    leaving(true)
{
    hp = 5;
    removeOnDie = false;

    //debug
//    hp = 25;
//    runSpeed = 10;
//    attackSpeed = 10;
//    bossKey = true;
}

Player::~Player()
{
    //dtor
}

bool Player::hurt(int dmg)
{
    if(hurtTimer > 0 || die)
    {
        return false;
    }
    hurtTimer = HURT_TIME;

    room->addParticle(new Particle(Particle::TYPE_WHITE_BLOOD, x+w/2, y+h/2));
    room->addParticle(new Particle(Particle::TYPE_WHITE_BLOOD, x+w/2, y+h/2));
    room->addParticle(new Particle(Particle::TYPE_WHITE_BLOOD, x+w/2, y+h/2));
    room->addParticle(new Particle(Particle::TYPE_WHITE_BLOOD, x+w/2, y+h/2));
    room->addParticle(new Particle(Particle::TYPE_WHITE_BLOOD, x+w/2, y+h/2));
    ResourceManager::playSound("hit.wav");
    return Entity::hurt(dmg);
}

void Player::step()
{
    if(die)
    {
        return;
    }
    if(fallingIntoHole)
    {
        if(fallTimer) fallTimer--;
        if(fallTimer == 0)
        {
            die = true;
        }
        return;
    }
    if(attackFrames) attackFrames--;
    if(hurtTimer > 0)
    {
        hurtTimer--;
    }
    float speed = sqrtf(dx * dx + dy * dy);
    if(speed > moveSpeed)
    {
        dx *= moveSpeed / speed;
        dy *= moveSpeed / speed;
    }

    moveSpeed = MOVE_SPEED;

    if(running)
    {
        //hacking some diminishing returns on run speed so you dont fall off map
        if(runSpeed <= 2)
        {
            moveSpeed += runSpeed * 0.5;
        }
        else if(runSpeed <= 5)
        {
            moveSpeed += runSpeed * 0.4;
        }
        else
        {
            moveSpeed += runSpeed * 0.35;
        }
        // add a trail
        if(moving)
        {
            room->addParticle(new Particle(Particle::TYPE_DUST, x + rand()%(int)(w), y + rand()%(int)(h)));
        }
    }


    if(attacking)
    {
        dx = 0;
        dy = 0;
    }

    x += dx;
    y += dy;

        dx *= 0.3;
        dy *= 0.3;

    if(dx > 0)
    {
        facingRight = true;
    }
    else if(dx < 0)
    {
        facingRight = false;
    }
    if(attacking && currSprite->isFinished())
    {
        attacking = false;
    }
    collideRoom();
    handleTransition();

    currSprite->step();
    moving = false;
}

void Player::draw(sf::RenderTarget& window)
{
    if(die)
    {
        if(fallingIntoHole)
        {
            currSprite = &fallSprite;
        }
        else
        {
            currSprite = &deadSprite;
        }
        currSprite->setColor(sf::Color::White);
    }
    else if(fabs(dx) > 0.2 || fabs(dy) > 0.2)
    {
        currSprite = &walkSprite;
    }
    else if(attacking)
    {
        currSprite = &attackSprite;
    }
    else
    {
        currSprite = &idleSprite;
    }
    if(hurtTimer > 0 && !die)
    {
        if(hurtTimer % 10 < 5)
        {
            currSprite->setColor(sf::Color::Transparent);
        }
        else
        {
            float amt = 1 - hurtTimer / (float)HURT_TIME;
            currSprite->setColor(sf::Color(155 * amt + 50, 155 * amt + 50, 155 * amt + 50));
        }

        currSprite->setPosition(x + w/2 + (rand()/(float)RAND_MAX - 0.5f), y + h/2);
    }
    else
    {
        currSprite->setColor(sf::Color::White);
        currSprite->setPosition(x + w/2, y + h/2);
    }
    if(fallingIntoHole)
    {
        float scale = (float)fallTimer/FALL_FRAMES;
        currSprite->setScale(scale);
    }
    else
    {
        currSprite->setScale(1.f);
    }
    if(descending)
    {
        currSprite = &fallSprite;
        if(leaving)
        {
            currSprite->setScale(1 - descentPercent);
            currSprite->setPosition(x + w/2, y + h/2);
        }
        else
        {
            currSprite->setScale(1);
            currSprite->setPosition(x + w/2, y + h/2 - (200.f - 200.f * (descentPercent)));
        }
    }
    currSprite->draw(window);
    currSprite->setFlipped(!facingRight);
}

void Player::drawStats(sf::RenderTarget& window, int px, int py)
{

    sf::Text text;
    text.setFont(ResourceManager::getDefaultFont());
    text.setColor(sf::Color::White);
    text.setCharacterSize(16);

    int *stat[5];
    stat[0] = &hp;
    stat[1] = &attackSpeed;
    stat[2] = &runSpeed;
    stat[3] = &numKeys;
    int bossKeys = (bossKey?1:0);
    stat[4] = &bossKeys; //wish I'd read this function before I made that a boolean~

    int row = 0;
    int col = 0;



    for (int i = 0; i < 5; i++)
    {
        statSprite.setFrame(i);
        int numStats = *stat[i];
        if(i > 0 && i < 3)
        {
            numStats--;
            //Exclude starting stats
        }
        for(int j = 0; j < numStats; j++)
        {
            statSprite.setPosition(px + col * 16, py + row * 16);
            statSprite.draw(window);
            col++;
            if(col > 4)
            {
                col = 0;
                row++;
            }
        }

    }
}

void Player::drawItemWave(sf::RenderTarget& window, float px, float py)
{
    static int frame = 0;
    frame++;
    //px, py is the center

    int *stat[3];
    stat[0] = &hp;
    stat[1] = &attackSpeed;
    stat[2] = &runSpeed;

    int col = 0;

    int totalStats = *stat[1]-1 + *stat[2] -1 + *stat[0];

    for (int i = 0; i < 3; i++)
    {
        statSprite.setFrame(i);
        int numStats = *stat[i];
        if(i > 0)
        {
            numStats--;
            //Exclude starting stats
        }
        for(int j = 0; j < numStats; j++)
        {
            statSprite.setPosition(px - (totalStats * 16 / 2)+ col * 16, cosf((frame + col) / 10.f) * 3 + py);
            statSprite.draw(window);
            col++;
        }
    }
}
void Player::doCollide(Entity* other)
{

}

void Player::handleTransition()
{
    transition = false;
    if(x+w < 0)
    {
        //trans left
        transitionDir = {-1, 0};
        transition = true;
    }
    if(x > room->w * Room::TILE_SIZE)
    {
        //trans right
        transitionDir = {1, 0};
        transition = true;
    }
    if(y+h < 0)
    {
        //trans up
        transitionDir = {0, -1};
        transition = true;
    }
    if(y > room->h * Room::TILE_SIZE)
    {
        //trans right
        transitionDir = {0, 1};
        transition = true;
    }
}

void Player::moveLeft()
{
    if(die) return;
    dx -= moveSpeed;
    direction.x = -1;
    direction.y = 0;
    moving = true;
}

void Player::moveRight()
{
    if(die) return;
    dx += moveSpeed;
    direction.x = 1;
    direction.y = 0;
    moving = true;
}

void Player::moveUp()
{
    if(die) return;
    dy -=moveSpeed;
    direction.x = 0;
    direction.y = -1;
    moving = true;
}

void Player::moveDown()
{
    if(die) return;
    dy += moveSpeed;
    direction.x = 0;
    direction.y = 1;
    moving = true;
}

void Player::attack()
{
    if(die) return;
    if(attackFrames) return;
    if(!attacking)
    {
        attacking = true;
        attackSprite.reset();
        sf::Vector2f pos;
        if(direction.x)
        {
            pos.x = direction.x > 0 ? (x+w) : (x-20);
            pos.y = (y+h/2)-(20/2);
        }
        else
        {
            pos.x = (x+w/2)-(20/2);
            pos.y = direction.y > 0 ? (y+h) : (y-20);
        }
        room->addEntity(new CombatParticle(pos.x, pos.y,
                                           CombatParticle::COMBAT_SLASH,
                                           sf::Vector2f(direction.x,direction.y),
                                           attackPow));
        attackFrames = max(1,(int)MAX_ATTACK_FRAMES - 3*attackSpeed);
        ResourceManager::playSound("attack.wav");
    }
}
void Player::startRun()
{
    if(die) return;
    running = true;
}
void Player::stopRun()
{
    running = false;
}

void Player::fallDownHole()
{
    if(fallingIntoHole)
        return;
    fallingIntoHole = true;
    fallTimer = FALL_FRAMES;
}
